(*
 * 名古屋大学大学院多元数理科学研究科 2015 年度連携大学院課外授業用資料
 * js_of_ocaml を利用したアニメーション運動量交換システムのサンプルプログラム
 *)

module Html = Dom_html
let jstr = Js.string
let document = Html.window##document

type stiff = INVALID | VALID of float
    
let n = ref 0
    
let ( +! ) (u1,u2) (v1,v2) =
  (u1 +. v1, u2 +. v2)

let ( *< ) c (u1, u2) =
  (c *. u1, c *. u2)

(* Runge-Kutta method *)
let runge_kutta_next (y, v) df h =
  let (ky1, kv1) = df (y, v) in
  let (ky2, kv2) = df ((y +! ((0.5*.h) *< ky1)), (v +! ((0.5*.h) *< kv1))) in
  let (ky3, kv3) = df ((y +! ((0.5*.h) *< ky2)), (v +! ((0.5*.h) *< kv2))) in
  let (ky4, kv4) = df ((y +! (h *< ky3)), (v +! (h *< kv3))) in
  (y +! ((h/.6.) *< (ky1 +! (2. *< ky2) +! (2. *< ky3) +! ky4)), v +! ((h/.6.) *< (kv1 +! (2. *< kv2) +! (2. *< kv3) +! kv4)))

let mb = 100.
let mu = 1.
let k_u = ref (VALID 20000.)
let g = 10.

let n = ref 0

let df ((yb,yu), (vb,vu)) =
  let k0 =
    match !k_u with
    | VALID k -> k
    | INVALID -> 0.
  in
  let dy = if yu -. yb < 30. then yu -. yb else 0. in
  let ab = -.(k0 /. mb) *. dy -. g in
  let au = if yu <= 320. then k0 /. mu *. dy -. g else 0. in
  ((vb,vu), (ab, au))


(* Canvas 作成 *)  
let create_canvas () =
  let c = Html.createCanvas document in
  c##width <- 320;
  c##height <- 320;
  c

let canvas = create_canvas ()


let int_input name value =
  let res = document##createDocumentFragment() in
  Dom.appendChild res (document##createTextNode(jstr name));
  let input = Html.createInput ~_type:(jstr "text") document in
  input##value <- jstr (string_of_int !value);
  input##onchange <- Html.handler
    (fun _ ->
       begin try
         value := int_of_string (Js.to_string (input##value))
       with Invalid_argument _ ->
         ()
       end;
       input##value <- jstr (string_of_int !value);
       Js._false);
  Dom.appendChild res input;
  res

let button name callback =
  let res = document##createDocumentFragment() in
  let input = Html.createInput ~_type:(jstr "submit") document in
  input##value <- jstr name;
  input##onclick <- Html.handler callback;
  Dom.appendChild res input;
  res

(* *)
let y_u = ref 320.
let v_u = ref (-30.)

let y_b = ref 280.
let v_b = ref (-30.)

let y_gnd = 320. +. !v_b *. 5. -. 0.5 *. g *. 5. *. 5.

let k_req = ref 20000

let tpi = 8. *. atan 1.
let nmax = 30
let theta n = tpi *. (float n) /. (float nmax)

let stop_flag = ref false


    
let draw ctx =
  ctx##fillStyle <- Js.string "#000066";
  ctx##fillRect (0., 0., 320., 320.);
  ctx##fillStyle <- Js.string "lightgray";
  ctx##strokeStyle <- Js.string "lightgray";
  match !k_u with
  | VALID _ ->
      begin
	ctx##fillRect (100., 320. -. !y_u, 10., 10.);
	ctx##beginPath ();
	let y_s = if !y_u -. !y_b <= 40.0 then !y_u -. !y_b -. 10. else 30. in
	let l_s = y_s /. 5. in
	ctx##moveTo (105., 320. -. (!y_b +. y_s));
	ctx##lineTo (110., 320. -. (!y_b +. 4. *. l_s));
	ctx##lineTo (100., 320. -. (!y_b +. 3. *. l_s));
	ctx##lineTo (110., 320. -. (!y_b +. 2. *. l_s));
	ctx##lineTo (100., 320. -. (!y_b +. l_s));
	ctx##lineTo (105., 320. -. !y_b);
	ctx##stroke ();
	
	ctx##fillRect (80., 320. -. !y_b, 50., 20.); 
	ctx##fillRect (80., 320. -. !y_b, 10., 40.); 
	ctx##fillRect (120., 320. -. !y_b, 10., 40.);
	ctx##fillStyle <- Js.string "darkgray";
	ctx##fillRect (0., 315., 320., 20.)
      end
  | INVALID ->
      begin
	ctx##fillRect (80., 320. -. !y_b, 50., 20.); 
	ctx##fillRect (80., 320. -. !y_b, 10., 40.); 
	ctx##fillRect (120., 320. -. !y_b, 10., 40.);
	ctx##fillStyle <- Js.string "darkgray";
	ctx##fillRect (0., 315., 320., 20.)
      end
    
let (>>=) = Lwt.bind


let rec loop canvas =
  let ctx = canvas##getContext (Html._2d_) in
  Lwt_js.sleep 0.01 >>= fun () ->
    ctx##clearRect (0., 0., float canvas##width, float canvas##height);
    draw ctx;
    let ((yb_next, yu_next), (vb_next, vu_next)) = runge_kutta_next ((!y_b, !y_u), (!v_b, !v_u)) df 0.01 in
    y_b := yb_next; y_u := yu_next;
    v_b := if !y_b <= y_gnd then -. 0.9 *. vb_next else vb_next;
    v_u := vu_next;
    n := !n + 1;
    loop canvas

let restart () =
  (* 初期条件 *)
  y_u := 320.;
  v_u := -30.;
  y_b := 280.;
  v_b := -30.;
  (* ばね定数 *)
  k_u := if 200 <= !k_req && !k_req <= 20000 then VALID (float_of_int !k_req) else INVALID;
  n := 0;
  ()
  

let onload _ =
  let main =
    Js.Opt.get (document##getElementById(jstr "main"))
      (fun () -> assert false)
  in
  Dom.appendChild main (int_input "剛性 / stiffness / raideur : " k_req);
  Dom.appendChild main (Html.createBr document);
  Dom.appendChild main
    (button " 再開 / restart / recommencer " (fun _ -> restart (); Js._false));
  canvas##width <- 320;
  canvas##height <- 320;
  Dom.appendChild Html.window##document##body canvas;
  canvas##getContext (Html._2d_) |> draw;
  ignore (loop canvas);
  Js._false

let () =
  Html.window##onload <- Html.handler onload
